from django.apps import AppConfig


class VerificationRedirectsConfig(AppConfig):
    name = 'rest_registration.contrib.verification_redirects'
    label = 'verification_redirects'
    verbose_name = 'Verification redirects'
